/*
 * Copyright (c) 2004 Bernd Walter <ticso@freebsd.org>
 *
 * $URL: https://devel.bwct.de/svn/projects/ubmb/tags/0.3/ubmb.c $
 * $Date: 2004-11-09 19:47:08 +0100 (Tue, 09 Nov 2004) $
 * $Author: ticso $
 * $Rev: 1797 $
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

/*
 * BWCT serial adapter driver
 */

#include <sys/cdefs.h>

#include <sys/param.h>
#include <sys/poll.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/bus.h>
#include <sys/ioccom.h>
#include <sys/fcntl.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/clist.h>
#include <sys/file.h>
#if __FreeBSD_version >= 502117 
#include <sys/module.h>
#endif

#if __FreeBSD_version >= 500014
#include <sys/selinfo.h>
#else
#include <sys/select.h>
#endif

#include <sys/sysctl.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#if __FreeBSD_version < 502117 
#include <dev/usb/usbdevs.h>
#endif

#include "ubmb.h"

/* XXX: remove befor commit */
#define USB_DEBUG

#ifdef USB_DEBUG
static int ubmbdebug = 0;
SYSCTL_NODE(_hw_usb, OID_AUTO, ubmb, CTLFLAG_RW, 0, "USB ubmb");
SYSCTL_INT(_hw_usb_ubmb, OID_AUTO, debug, CTLFLAG_RW,
	   &ubmbdebug, 0, "ubmb debug level");
#define DPRINTF(x)      do { \
				if (ubmbdebug) \
					logprintf x; \
			} while (0)

#define DPRINTFN(n, x)  do { \
				if (ubmbdebug > (n)) \
					logprintf x; \
			} while (0)
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

#define ISSET(t, f)	((t) & (f))
#define SET(t, f)	(t) |= (f)
#define CLR(t, f)	(t) &= ~((unsigned)(f))

struct ubmb_softc {
	USBBASEDEVICE		sc_dev;
	usbd_device_handle	sc_udev;
	usbd_interface_handle	sc_iface;	/* data interface */
	int			sc_ifaceno;

	int			sc_refcnt;
	u_char			sc_dying;
	u_char			sc_isopen;

	u_char			sc_bulkin_size;
	int			sc_bulkin_no;	/* bulk in endpoint address */
	usbd_pipe_handle	sc_bulkin_pipe;	/* bulk in pipe */

	u_char			sc_bulkout_size;
	int			sc_bulkout_no;	/* bulk out endpoint address */
	usbd_pipe_handle	sc_bulkout_pipe;/* bulk out pipe */
#if __FreeBSD_version >= 502117 
	struct cdev		*dev;
#else
	dev_t			dev;
#endif
};

Static d_open_t  ubmb_open;
Static d_close_t ubmb_close;
Static d_read_t  ubmb_read;
Static d_write_t ubmb_write;
Static d_ioctl_t ubmb_ioctl;
Static d_poll_t ubmb_poll;

static struct cdevsw ubmb_cdevsw = {
#if __FreeBSD_version > 502102
	.d_version = 	D_VERSION,
#endif
	.d_open =       ubmb_open,
	.d_close =      ubmb_close,
	.d_read =       ubmb_read,
	.d_write =      ubmb_write,
	.d_ioctl =      ubmb_ioctl,
	.d_poll =       ubmb_poll,
	.d_name =       "ubmb",
#if __FreeBSD_version > 502102
	.d_flags =      D_NEEDGIANT,
#endif
#if __FreeBSD_version < 500014
	.d_bmaj =       -1,
#endif
};

USB_DECLARE_DRIVER(ubmb);

USB_MATCH(ubmb)
{
	USB_MATCH_START(ubmb, uaa);
	usb_string_descriptor_t us;
	usb_interface_descriptor_t *id;
	usb_device_descriptor_t *dd;
	int err;
#if __FreeBSD_version >= 502117 
	int size;
#endif

	if (uaa->iface == NULL)
		return (UMATCH_NONE);

	DPRINTFN(20,("ubmb: vendor=0x%x, product=0x%x\n",
		     uaa->vendor, uaa->product));

	dd = usbd_get_device_descriptor(uaa->device);
	if (dd == NULL) {
		printf("ubmb: failed to get device descriptor\n");
		return (UMATCH_NONE);
	}

	id = usbd_get_interface_descriptor(uaa->iface);
	if (id == NULL) {
		printf("ubmb: failed to get interface descriptor\n");
		return (UMATCH_NONE);
	}

#if __FreeBSD_version >= 502117 
	err = usbd_get_string_desc(uaa->device, dd->iManufacturer, 0, &us,
	    &size);
#else
	err = usbd_get_string_desc(uaa->device, dd->iManufacturer, 0, &us);
#endif
	if (err != 0)
		return (UMATCH_NONE);

	/* check if this is a BWCT vendor specific ubmb interface */
	if (strcmp((char*)us.bString, "B\0W\0C\0T\0") == 0 &&
	    id->bInterfaceClass == 0xff && id->bInterfaceSubClass == 0x02)
		return (UMATCH_VENDOR_IFACESUBCLASS);

	return (UMATCH_NONE);
}

USB_ATTACH(ubmb)
{
	USB_ATTACH_START(ubmb, sc, uaa);
	usbd_device_handle udev = uaa->device;
	char *devinfo;
	usb_endpoint_descriptor_t *ed;
	usbd_status err;
	int i;
	uint8_t epcount;

	devinfo = malloc(1024, M_USBDEV, M_WAITOK);
	usbd_devinfo(udev, 0, devinfo);
	USB_ATTACH_SETUP;
#if __FreeBSD_version < 502117 
	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), devinfo);
#endif

	DPRINTFN(10,("\nubmb_attach: sc=%p\n", sc));

	sc->sc_udev = udev = uaa->device;
	sc->sc_iface = uaa->iface;

	/* find our bulk endpoints */
	epcount = 0;
	usbd_endpoint_count(sc->sc_iface, &epcount);
	sc->sc_bulkin_no = -1;
	sc->sc_bulkout_no = -1;
	for (i = 0; i < epcount; i++) {
		ed = usbd_interface2endpoint_descriptor(sc->sc_iface, i);
		if (ed == NULL) {
			printf("%s: couldn't get ep %d\n",
			USBDEVNAME(sc->sc_dev), i);
			USB_ATTACH_ERROR_RETURN;
		}
		if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
		    UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			sc->sc_bulkin_no = ed->bEndpointAddress;
			sc->sc_bulkin_size = UGETW(ed->wMaxPacketSize);
		} else if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_OUT &&
		    UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			sc->sc_bulkout_no = ed->bEndpointAddress;
			sc->sc_bulkout_size = UGETW(ed->wMaxPacketSize);
		}
	}
	if (sc->sc_bulkin_no == -1) {
		printf("%s: could not find bulk in endpoint\n",
		    USBDEVNAME(sc->sc_dev));
		sc->sc_dying = 1;
		USB_ATTACH_ERROR_RETURN;
	}
	if (sc->sc_bulkout_no == -1) {
		printf("%s: could not find bulk out endpoint\n",
		USBDEVNAME(sc->sc_dev));
		sc->sc_dying = 1;
		USB_ATTACH_ERROR_RETURN;
	}

	/* Open the bulk pipes */
	/* Bulk-in pipe */
	err = usbd_open_pipe(sc->sc_iface, sc->sc_bulkin_no, 0,
	    &sc->sc_bulkin_pipe);
	if (err) {
		printf("%s: open bulk in error (addr %d): %s\n",
		USBDEVNAME(sc->sc_dev), sc->sc_bulkin_no,
		usbd_errstr(err));
		goto bad;
	}
	/* Bulk-out pipe */
	err = usbd_open_pipe(sc->sc_iface, sc->sc_bulkout_no,
	    USBD_EXCLUSIVE_USE, &sc->sc_bulkout_pipe);
	if (err) {
		printf("%s: open bulk out error (addr %d): %s\n",
		USBDEVNAME(sc->sc_dev), sc->sc_bulkout_no,
		usbd_errstr(err));
		goto fail_1;
	}

	sc->dev = make_dev(&ubmb_cdevsw, USBDEVUNIT(sc->sc_dev), UID_ROOT,
	    GID_WHEEL, 0660, "%s", USBDEVNAME(sc->sc_dev));

	free(devinfo, M_USBDEV);
	USB_ATTACH_SUCCESS_RETURN;

fail_1:
	usbd_close_pipe(sc->sc_bulkin_pipe);
	sc->sc_bulkin_pipe = NULL;
bad:
	DPRINTF(("ubmb_attach: ATTACH ERROR\n"));
	free(devinfo, M_USBDEV);

	USB_ATTACH_ERROR_RETURN;
}

USB_DETACH(ubmb)
{
	int s;

	USB_DETACH_START(ubmb, sc);

	DPRINTF(("ubmb_detach: sc=%p\n", sc));

	sc->sc_dying = 1;

	if (sc->sc_bulkin_pipe != NULL)
		usbd_abort_pipe(sc->sc_bulkin_pipe);
	if (sc->sc_bulkout_pipe != NULL)
		usbd_abort_pipe(sc->sc_bulkout_pipe);

	s = splusb();
	if (--sc->sc_refcnt >= 0) {
		/* Wait for processes to go away. */
		usb_detach_wait(USBDEV(sc->sc_dev));
	}
	splx(s);

	destroy_dev(sc->dev);

	return (0);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_open(struct cdev *dev, int flag, int mode, usb_proc_ptr p)
#else
ubmb_open(dev_t dev, int flag, int mode, usb_proc_ptr p)
#endif
{
	struct ubmb_softc *sc;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);
	
	if (sc->sc_isopen)
		return (EBUSY);

	sc->sc_isopen = 1;

	return (0);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_close(struct cdev *dev, int flag, int mode, usb_proc_ptr p)
#else
ubmb_close(dev_t dev, int flag, int mode, usb_proc_ptr p)
#endif
{
	struct ubmb_softc *sc;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);

	sc->sc_isopen = 0;

	return (0);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flag, usb_proc_ptr p)
#else
ubmb_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, usb_proc_ptr p)
#endif
{
	struct ubmb_softc *sc;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);

	return (EINVAL);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_read(struct cdev *dev, struct uio *uio, int flag)
#else
ubmb_read(dev_t dev, struct uio *uio, int flag)
#endif
{
	struct ubmb_softc *sc;
	usbd_xfer_handle xfer;
	usbd_status err;
	int error = 0;
	uint32_t n;
	char *buf;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);

	sc->sc_refcnt++;

	buf = malloc(sc->sc_bulkin_size, M_USBDEV, M_WAITOK);

	xfer = usbd_alloc_xfer(sc->sc_udev);
	if (xfer == 0) {
		error = EIO;
		goto bad;
	}

	/* TODO: limit to not overflow uio */
	for (;;) {
		n = sc->sc_bulkin_size;
		DPRINTFN(1, ("ubmbread: transfer read %d bytes\n", n));
		err = usbd_bulk_transfer(xfer, sc->sc_bulkin_pipe,
		    USBD_SHORT_XFER_OK,
		    USBD_DEFAULT_TIMEOUT, buf, &n, "ubmbrb");
		if (err) {
			if (err == USBD_INTERRUPTED)
				error = EINTR;
			else if (err == USBD_TIMEOUT)
				error = ETIMEDOUT;
			else
				error = EIO;
			break;
		}
		DPRINTFN(2, ("ubmbread: transfer received %d bytes\n", n));
		if (n != 0)
			error = uiomove(buf, n, uio);
			if (error)
				break;
		if (n < sc->sc_bulkin_size)
			break;
	}


	usbd_free_xfer(xfer);
bad:
	free(buf, M_USBDEV);

	if (--sc->sc_refcnt < 0)
		usb_detach_wakeup(USBDEV(sc->sc_dev));

	return (error);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_write(struct cdev *dev, struct uio *uio, int flag)
#else
ubmb_write(dev_t dev, struct uio *uio, int flag)
#endif
{
	struct ubmb_softc *sc;
	usbd_xfer_handle xfer;
	usbd_status err;
	int error = 0;
	uint32_t n;
	char *buf;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);

	sc->sc_refcnt++;

	buf = malloc(uio->uio_resid, M_USBDEV, M_WAITOK);

	xfer = usbd_alloc_xfer(sc->sc_udev);
	if (xfer == 0) {
		error = EIO;
		goto bad2;
	}
	n = uio->uio_resid;
	error = uiomove(buf, n, uio);
	if (error) {
		goto bad1;
	}
	DPRINTFN(1, ("ubmbwrite: transfer write %d bytes\n", n));
	err = usbd_bulk_transfer(xfer, sc->sc_bulkout_pipe, USBD_FORCE_SHORT_XFER,
	    USBD_DEFAULT_TIMEOUT, buf, &n, "ubmbwb");
	if (err) {
		if (err == USBD_INTERRUPTED)
			error = EINTR;
		else if (err == USBD_TIMEOUT)
			error = ETIMEDOUT;
		else
			error = EIO;
	} else {
		DPRINTFN(2, ("ubmbwrite: transfer send %d bytes\n", n));
	}

bad1:
	usbd_free_xfer(xfer);
bad2:
	free(buf, M_USBDEV);

	if (--sc->sc_refcnt < 0)
		usb_detach_wakeup(USBDEV(sc->sc_dev));
	return (error);
}

static int
#if __FreeBSD_version >= 502117 
ubmb_poll(struct cdev *dev, int events, usb_proc_ptr p)
#else
ubmb_poll(dev_t dev, int events, usb_proc_ptr p)
#endif
{
	struct ubmb_softc *sc;
	int revents = 0;

	USB_GET_SC(ubmb, dev2unit(dev), sc);

	if (sc->sc_dying)
		return (EIO);

	/* TODO */
	revents |= events & (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM);
	return (revents);
}

DRIVER_MODULE(ubmb, uhub, ubmb_driver, ubmb_devclass, usbd_driver_load, 0);

